/*
 * JOnAS: Java(TM) Open Application Server
 * Copyright (C) 1999 Bull S.A.
 * Contact: jonas-team@objectweb.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: JEntitySwitchCRU.java,v 1.1.2.1 2004/02/02 09:51:22 durieuxp Exp $
 * --------------------------------------------------------------------------
 */

package org.objectweb.jonas_ejb.container;

import java.rmi.NoSuchObjectException;
import java.rmi.RemoteException;
import java.util.HashMap;
import javax.ejb.EJBException;
import javax.ejb.EntityBean;
import javax.ejb.ObjectNotFoundException;
import javax.ejb.TimedObject;
import javax.ejb.Timer;
import javax.ejb.TimerService;
import javax.ejb.TransactionRolledbackLocalException;
import javax.transaction.RollbackException;
import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.Transaction;
import org.objectweb.jonas_ejb.deployment.api.EntityDesc;
import org.objectweb.jonas_ejb.deployment.api.MethodDesc;
import org.objectweb.util.monolog.api.BasicLevel;

/**
 * JEntitySwitch is used internally to synchronize accesses to the entity context
 * and thus to the entity bean instance.
 * All parts common to EJBObject and EJBLocalObject should be here.
 * Different policies can be applied to manage context/instance pairs:
 * - only 1 pair (container manages the transaction isolation)
 * - 1 pair for each transaction (transaction isolation managed by DataBase)
 * - 2 pairs (1 for transactional accesses, 1 for non transaction accesses)
 * @author Philippe Durieux
 * @author Philippe Coq
 */
public class JEntitySwitchCRU extends JEntitySwitch {

    /**
     * unique EntityContext
     */
    protected JEntityContext itContext = null;

    /**
     * empty constructor. Object is initialized via init() because it is
     * implemented differently according to jorm mappers.
     */
    public JEntitySwitchCRU() {
        lockpolicy = EntityDesc.LOCK_CONTAINER_READ_UNCOMMITTED;
        txUpdates = true;
    }

    protected void initpolicy(JEntityFactory bf) {
        lazyregister = true;
    }

    protected JEntityContext getContext4Tx(Transaction tx) {
        return itContext;
    }

    protected void setContext4Tx(Transaction tx, JEntityContext ctx) {
        itContext = ctx;
    }

    protected void removeContext4Tx(Transaction tx) {
        discarded = false;
        itContext = null;
    }

    /**
     * Map a context and its instance.
     * @param tx - the Transaction object
     * @param bctx - the JEntityContext to bind if not null
     * @param forced - force to take this context. (case of create)
     * @param holdit - increment count to hold it, a release will be called later.
     * @return JEntityContext actually mapped
     */
    public synchronized JEntityContext mapICtx(Transaction tx, JEntityContext bctx, boolean forced, boolean holdit) {

        // Synchronization.
        if (tx == null) {
            // ------------- IH begin -------------
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH");
        } else {
            // ------------- IT begin -------------
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IT");
        }

        // Choose the context to use.
        boolean newtrans = false;
        JEntityContext jec = itContext;
        if (forced) {
            // If the new context is enforced, we must first release the older
            if (jec != null) {
                if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "new context is enforced!");
                bf.releaseJContext(jec);
                itContext = null;
                discarded = false;
            }
            jec = bctx;
            itContext = jec;
            jec.initEntityContext(this);
            newtrans = true;
        } else {
            if (jec != null) {
                // Reuse the Context for this transaction.
                // If a context was supplied, release it first.
                if (bctx != null) {
                    if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "a context was supplied!");
                    bf.releaseJContext(bctx);
                }
                jec.reuseEntityContext();
                newtrans = true; // TODO: not always a new tx (if more than 1 modifying call on this)
            } else {
                if (bctx != null) {
                    jec = bctx;
                } else {
                    // no Context available : get one from the pool.
                    jec = (JEntityContext) bf.getJContext();
                }
                jec.initEntityContext(this);
                jec.activate(true);
                itContext = jec; // after activate
                newtrans = true;
            }
        }

        if (tx == null) {
            if (holdit) {
                countIH++;
                if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "mapICtx IH count=" + countIH);
                if (shared && countIH == 1) {
                    // reload state that could have been modified by transactions.
                    jec.activate(false);
                }
            }
        }

        return jec;
    }

    /**
     * Release a context/instance at end of request.
     * @param tx - transaction associated to this context
     */
    public synchronized void releaseICtx(Transaction tx) {

        if (tx == null) {
            // ------------- IH end -------------
            JEntityContext jec = itContext;
            countIH--;
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "releaseICtx IH count=" + countIH);
            if (countIH == 0) {
                if (discarded) {
                    if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "discarded!");
                    bf.releaseJContext(jec);
                    itContext = null;
                    discarded = false;
                } else {
                    if (jec.isMarkedRemoved()) {
                        if (TraceEjb.isDebugContext()) TraceEjb.context.log(BasicLevel.DEBUG, ident + "Marked Removed!");
                        bf.releaseJContext(jec);
                        itContext = null;
                        noLongerUsed();
                    }
                }
                if (waiters > 0) {
                    TraceEjb.synchro.log(BasicLevel.DEBUG, ident + " notify");
                    notifyAll();
                }
            }
        } else {
            // ------------- IT end -------------
            // nothing to do now. Wait for tx completed.
            if (TraceEjb.isDebugSwapper()) TraceEjb.swapper.log(BasicLevel.DEBUG,ident + "releaseICtx IT");

        }
    }

    public boolean passivateIH(boolean passivation) {  
        return true;            // NEVER
    }

    public void endIH() {
        return;           // NEVER
    }

    /**
     * This transaction has just modified this instance. (CMP2 only)
     * @param tx transaction
     */
    public synchronized void notifyWriting(Transaction tx, JEntityContext bctx) {
        TraceEjb.swapper.log(BasicLevel.DEBUG, ident);
        if (writingtx != null) {
            TraceEjb.logger.log(BasicLevel.WARN, "Conflict: 2 transactions attempt to write the same instance");
            try {
                tx.setRollbackOnly();
            } catch (SystemException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot set current transaction RollbackOnly");
            }
            // Since we share the same instance, we must rollback also
            // the current transaction (else data would be corrupted).
            try {
                writingtx.setRollbackOnly();
            } catch (SystemException e) {
                TraceEjb.logger.log(BasicLevel.ERROR, "cannot set other transaction RollbackOnly");
            }
            // We should also avoid modifying instance, in case writingtx is
            // already committing (setRollbackOnly will fail)
            throw new EJBException("Conflict writing entity bean");
        } else {
            // Instance must be written later.
            registerCtx(tx, bctx);
            writingtx = tx;
        }
    }

}
